from django.contrib.auth.decorators import user_passes_test
from django.core.exceptions import PermissionDenied

from .utils import get_default_error_message, get_permission_full_name


def permission_required(perm: str, login_url=None, raise_exception=True, message=None):
    """
    Decorator for views that checks whether a user has a particular permission
    enabled, redirecting to the log-in page if necessary.
    If the raise_exception parameter is given the PermissionDenied exception
    is raised.

    if the permission is located in `plain_permissions` app, app_name ('plain_permissions') can be omitted

    i.e. if permission in `plain_permissions` app is `activate_account`, the following will work:
    @permission_required('activate_account')
    @permission_required('plain_permissions.activate_account')

    if the permission is located in another app, the app_name must be included
    @permission_required('app_name.permission_name')
    @permission_required('orders.Order.delete_order')

    """

    def check_perms(user):
        if isinstance(perm, str):
            perms = (get_permission_full_name(perm),)
        else:
            perms = perm
        # First check if the user has the permission (even anon users)
        if user.has_perms(perms):
            return True
        # In case the 403 handler should be called raise the exception
        error_message = message
        if error_message is None:
            error_message = get_default_error_message(perm)
        if raise_exception:
            raise PermissionDenied(error_message)
        # As the last resort, show the login form
        return False

    return user_passes_test(check_perms, login_url=login_url)



from django.contrib.auth.models import Permission, User
from django.core.exceptions import PermissionDenied
from django.test import TestCase

from plain_permissions.utils import check_permission


class TestUtils(TestCase):
    def test_check_permission_should_not_raise_if_user_has_permission(self):
        user = User.objects.create_user(username="test", password="test")
        print(Permission.objects.values_list("content_type__app_label"))
        perm = Permission.objects.get(
            content_type__app_label="plain_permissions", content_type__model="permission", codename="disable_user"
        )
        user.user_permissions.add(perm)
        check_permission(user, "disable_user")

    def test_check_permission_should_raise_if_user_does_not_have_permission(self):
        user = User.objects.create_user(username="test", password="test")
        with self.assertRaises(PermissionDenied):
            check_permission(user, "disable_user")

    def test_has_perm_should_return_true_if_user_has_permission(self):
        user = User.objects.create_user(username="test", password="test")
        perm = Permission.objects.get(
            content_type__app_label="plain_permissions", content_type__model="permission", codename="disable_user"
        )
        user.user_permissions.add(perm)
        self.assertTrue(user.has_perm("plain_permissions.disable_user"))

    def test_has_perm_should_return_false_if_user_does_not_have_permission(self):
        user = User.objects.create_user(username="test", password="test")
        self.assertFalse(user.has_perm("plain_permissions.disable_user"))



from django.contrib.auth.models import Permission
from django.db import models


def permission_string_method(self):
    return f"{self.name}"


Permission.__str__ = permission_string_method


class Permission(models.Model):
    """Fake model to attach the permissions to it"""

    class Meta:
        managed = False
        db_table = ""
        default_permissions = []



from django.apps import AppConfig
from django.db.models.signals import post_migrate


class PermissionsConfig(AppConfig):
    name = "plain_permissions"

    def ready(self):
        from plain_permissions import permissions_settings
        from plain_permissions.signal_handlers import sync_permissions

        if permissions_settings.SYNC_PERMISSIONS_POST_MIGRATE:
            post_migrate.connect(sync_permissions, sender=self)

        if permissions_settings.MONKEYPATCH_USER:
            from django.contrib.auth import get_user_model

            from plain_permissions.utils import custom_has_perm

            User = get_user_model()
            User.has_perm = custom_has_perm



from __future__ import annotations

from typing import TypedDict

from django.conf import settings


class PermissionsSettingsDefaults(TypedDict):
    PERMISSIONS: list[tuple[str, str]]
    DEFAULT_ERROR_MESSAGE: str
    MONKEYPATCH_USER: bool
    OVERRIDE_GROUP_ADMIN: bool
    OVERRIDE_USER_ADMIN: bool
    SYNC_PERMISSIONS_POST_MIGRATE: bool


DEFAULTS: PermissionsSettingsDefaults = {
    "PERMISSIONS": [],
    "DEFAULT_ERROR_MESSAGE": 'You do not have permission to "%s"',
    "MONKEYPATCH_USER": True,
    "OVERRIDE_GROUP_ADMIN": True,
    "OVERRIDE_USER_ADMIN": False,
    "SYNC_PERMISSIONS_POST_MIGRATE": True,
}


class PermissionsSettings:
    def __init__(self, user_settings=None, defaults=None):
        if user_settings:
            self._user_settings = user_settings
        self.defaults = defaults or DEFAULTS

    @property
    def user_settings(self):
        if not hasattr(self, "_user_settings"):
            self._user_settings = getattr(settings, "PERMISSIONS_SETTINGS", {})
        return self._user_settings

    def __getattr__(self, attr):
        if attr not in self.defaults:
            raise AttributeError("Invalid Permissions setting: '%s'" % attr)

        try:
            # Check if present in user settings
            val = self.user_settings[attr]
        except KeyError:
            # Fall back to defaults
            val = self.defaults[attr]

        setattr(self, attr, val)
        return val


permissions_settings = PermissionsSettings(None, DEFAULTS)



from django.conf import settings
from django.contrib.auth.models import Permission
from django.contrib.contenttypes.models import ContentType
from django.test import TestCase

from plain_permissions.signal_handlers import sync_permissions
from tests.models import Post


class TestSyncPermissions(TestCase):
    def test_sync_permissions(self):
        # act
        # sync_permissions(None) is called after migration

        # assert
        from plain_permissions.models import Permission as CustomPermission

        content_type = ContentType.objects.get_for_model(CustomPermission)

        for perm in ["merge_customers", "give_edit_permission", "disable_user"]:
            assert Permission.objects.filter(content_type=content_type, codename=perm).exists()

        for perm in ["delete_post"]:
            assert Permission.objects.filter(
                content_type=ContentType.objects.get_for_model(Post), codename=perm
            ).exists()

    def test_removing_permission(self):
        # arrange
        from plain_permissions import permissions_settings

        settings.PERMISSIONS_SETTINGS["PERMISSIONS"] = [
            ("merge_customers", "Merge Customers"),
        ]
        del permissions_settings.PERMISSIONS
        # act
        sync_permissions(None)

        # assert
        from plain_permissions.models import Permission as CustomPermission

        content_type = ContentType.objects.get_for_model(CustomPermission)

        for perm in ["give_edit_permission", "disable_user"]:
            assert False == Permission.objects.filter(content_type=content_type, codename=perm).exists()



from __future__ import annotations

from functools import reduce
from operator import or_

from django.contrib.auth.models import Permission, _user_has_perm
from django.core.exceptions import PermissionDenied
from django.db.models import Q

from plain_permissions import permissions_settings
from plain_permissions.apps import PermissionsConfig


def check_permission(user, perm: str) -> None:
    """Check if the user has the permission, if not raise PermissionDenied"""
    if not user.has_perm(get_permission_full_name(perm)):
        raise PermissionDenied(get_default_error_message(perm))


def get_default_error_message(perm: str) -> str:
    from django.contrib.auth.models import Permission
    from django.contrib.contenttypes.models import ContentType

    from plain_permissions.models import Permission as CustomPermission

    content_type = ContentType.objects.get_for_model(CustomPermission)
    permission_readable_name = Permission.objects.get(content_type=content_type, codename=perm).name
    error_message = permissions_settings.DEFAULT_ERROR_MESSAGE % permission_readable_name
    return error_message


def get_permission_full_name(perm: str) -> str:
    """if app_name is not included in the permission name, it will default to plain_permissions app_name"""
    my_perm = perm
    if "." not in my_perm:
        my_perm = PermissionsConfig.name + "." + perm
    return my_perm


def get_permissions_queryset():
    """return `Permission Queryset` that are defined in settings.PERMISSIONS_SETTINGS and in the database
    ignores the builtin permissions that are not mentioned in settings.PERMISSIONS_SETTINGS
    """
    custom_permissions = Q(content_type__app_label="plain_permissions", content_type__model="permission")
    q_objs = [custom_permissions]
    builtin_permissions = [p[0] for p in get_permissions_tuple() if "." in p[0]]
    for p in builtin_permissions:
        app_label, model_name, permission_name = p.split(".")
        q_objs.append(
            Q(codename=permission_name, content_type__app_label=app_label, content_type__model=model_name.lower())
        )
    query = reduce(or_, q_objs)
    return Permission.objects.filter(query)


def custom_has_perm(user, perm: str, obj=None) -> bool:
    """
    Return True if the user has the specified permission. Query all
    available auth backends, but return immediately if any backend returns
    True. Thus, a user who has permission from a single auth backend is
    assumed to have permission in general. If an object is provided, check
    permissions for that object.
    """
    print("MONKEYYYYYYYYYYYYY")
    perm = get_permission_full_name(perm)

    # Active superusers have all permissions.
    if user.is_active and user.is_superuser:
        return True

    # Otherwise we need to check the backends.
    return _user_has_perm(user, perm, obj)


def get_permissions_tuple() -> list[tuple[str, str]]:
    return permissions_settings.PERMISSIONS



INSTALLED_APPS = (
    "django.contrib.auth",
    "django.contrib.contenttypes",
    "tests",
    "plain_permissions",
)
DATABASES = {
    "default": {
        "ENGINE": "django.db.backends.sqlite3",
        "NAME": ":memory:",
    },
}
SECRET_KEY = "dummy"

CACHES = {
    "default": {
        "BACKEND": "django.core.cache.backends.locmem.LocMemCache",
    }
}


DEFAULT_AUTO_FIELD = "django.db.models.AutoField"

USE_TZ = False

PERMISSIONS_SETTINGS = {
    "PERMISSIONS": [
        ("tests.Post.delete_post", "Delete Posts"),  # reuse built-in permission
        ("merge_customers", "Merge Customers"),  # custom permission
        ("give_edit_permission", "Give Permission To Edit Permissions"),
        ("disable_user", "Disable Users"),
    ],
    "DEFAULT_ERROR_MESSAGE": 'You do not have permission to "%s".',
    "MONKEYPATCH_USER": True,
    "OVERRIDE_GROUP_ADMIN": True,
    "OVERRIDE_USER_ADMIN": True,
    "SYNC_PERMISSIONS_POST_MIGRATE": True,
}



import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):
    initial = True

    operations = [
        migrations.CreateModel(
            name="Author",
            fields=[
                ("id", models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("first_name", models.CharField(max_length=200)),
                ("last_name", models.CharField(max_length=200)),
            ],
        ),
        migrations.CreateModel(
            name="Post",
            fields=[
                ("id", models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("title", models.CharField(max_length=255)),
                ("publish_date", models.DateField()),
                (
                    "author",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE, related_name="posts", to="tests.author"
                    ),
                ),
            ],
        ),
    ]



""" override default setting by adding `PERMISSIONS_SETTINGS` to your settings.py
then run manage.py migrate or manage.py sync_permissions to sync your PERMISSIONS
"""

__version__ = "0.0.1"


from plain_permissions.settings import permissions_settings



from django import forms
from django.contrib import admin
from django.contrib.admin import widgets
from django.contrib.auth import get_user_model
from django.contrib.auth.admin import UserAdmin
from django.contrib.auth.forms import UserChangeForm
from django.contrib.auth.models import Group
from django.utils.translation import gettext_lazy as _

from plain_permissions.settings import permissions_settings
from plain_permissions.utils import get_permissions_queryset


class PlainPermissionsGroupAdminForm(forms.ModelForm):
    class Meta:
        model = Group
        exclude = []

    permissions = forms.ModelMultipleChoiceField(
        get_permissions_queryset(), widget=widgets.FilteredSelectMultiple(_("permissions"), False)
    )


class PlainPermissionsGroupAdmin(admin.ModelAdmin):
    form = PlainPermissionsGroupAdminForm
    search_fields = ("name",)
    ordering = ("name",)


class PlainPermissionsUserChangeForm(UserChangeForm):
    user_permissions = forms.ModelMultipleChoiceField(
        get_permissions_queryset(),
        required=False,
        widget=widgets.FilteredSelectMultiple(_("user_permissions"), False),
        help_text='Hold down "Control", or "Command" on a Mac, to select more than one.',
    )


class MyUserAdmin(UserAdmin):
    form = PlainPermissionsUserChangeForm


if permissions_settings.OVERRIDE_GROUP_ADMIN:
    admin.site.unregister(Group)
    admin.site.register(Group, PlainPermissionsGroupAdmin)
if permissions_settings.OVERRIDE_USER_ADMIN:
    admin.site.unregister(get_user_model())
    admin.site.register(get_user_model(), MyUserAdmin)



from os import path

from setuptools import setup


this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, "README.md"), encoding="utf-8") as f:
    long_description = f.read()
if __name__ == "__main__":
    setup(long_description=long_description, long_description_content_type="text/markdown")



from django.contrib.auth.models import Permission
from django.contrib.contenttypes.models import ContentType

from plain_permissions.models import Permission as CustomPermission
from plain_permissions.utils import get_permissions_tuple


def sync_permissions(sender, **kwargs):
    populate_permissions()
    delete_missing_permissions()


def populate_permissions():
    """add any permission that is not already in the database and update their name if they exists
    will skip permission that are in other apps.
    permissions that are in other apps have '.' in it i.e. 'app_name.model_name.permission_name'
    """
    for code_name, name in get_permissions_tuple():
        if "." in code_name:
            app_label, model_name, code_name = code_name.split(".")
            content_type = ContentType.objects.get(app_label=app_label, model=model_name.lower())
        else:
            content_type = ContentType.objects.get_for_model(CustomPermission)
        permission, _ = Permission.objects.update_or_create(
            codename=code_name, content_type=content_type, defaults={"name": name}
        )


def delete_missing_permissions():
    """delete all permissions in located in this app, but not in PERMISSIONS"""
    content_type = ContentType.objects.get_for_model(CustomPermission)
    code_names = [code_name for (code_name, name) in get_permissions_tuple() if "." not in code_name]
    Permission.objects.filter(content_type=content_type).exclude(codename__in=code_names).delete()



from django.db import models


class Author(models.Model):
    first_name = models.CharField(max_length=200)
    last_name = models.CharField(max_length=200)


class Post(models.Model):
    title = models.CharField(max_length=255)
    publish_date = models.DateField()
    author = models.ForeignKey(Author, on_delete=models.CASCADE, related_name="posts")
